use super::GetStandardProperties;
use super::StandardProperties;

// TODO: Why did we make Object implement PartialEq again? Was it just for tests?
#[derive(Debug, PartialEq)]
pub enum Object<'s> {
    Bold(Bold<'s>),
    Italic(Italic<'s>),
    Underline(Underline<'s>),
    StrikeThrough(StrikeThrough<'s>),
    Code(Code<'s>),
    Verbatim(Verbatim<'s>),
    PlainText(PlainText<'s>),
    RegularLink(RegularLink<'s>),
    RadioLink(RadioLink<'s>),
    RadioTarget(RadioTarget<'s>),
    PlainLink(PlainLink<'s>),
    AngleLink(AngleLink<'s>),
    OrgMacro(OrgMacro<'s>),
    Entity(Entity<'s>),
    LatexFragment(LatexFragment<'s>),
    ExportSnippet(ExportSnippet<'s>),
    FootnoteReference(FootnoteReference<'s>),
    Citation(Citation<'s>),
    CitationReference(CitationReference<'s>),
    InlineBabelCall(InlineBabelCall<'s>),
    InlineSourceBlock(InlineSourceBlock<'s>),
    LineBreak(LineBreak<'s>),
    Target(Target<'s>),
    StatisticsCookie(StatisticsCookie<'s>),
    Subscript(Subscript<'s>),
    Superscript(Superscript<'s>),
    Timestamp(Timestamp<'s>),
}

#[derive(Debug, PartialEq)]
pub struct Bold<'s> {
    pub source: &'s str,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug, PartialEq)]
pub struct Italic<'s> {
    pub source: &'s str,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug, PartialEq)]
pub struct Underline<'s> {
    pub source: &'s str,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug, PartialEq)]
pub struct StrikeThrough<'s> {
    pub source: &'s str,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug, PartialEq)]
pub struct Code<'s> {
    pub source: &'s str,
    pub contents: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct Verbatim<'s> {
    pub source: &'s str,
    pub contents: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct PlainText<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct RegularLink<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct RadioTarget<'s> {
    pub source: &'s str,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug, PartialEq)]
pub struct RadioLink<'s> {
    pub source: &'s str,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug, PartialEq)]
pub struct PlainLink<'s> {
    pub source: &'s str,
    pub link_type: &'s str,
    pub path: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct AngleLink<'s> {
    pub source: &'s str,
    pub link_type: &'s str,
    pub path: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct OrgMacro<'s> {
    pub source: &'s str,
    pub macro_name: &'s str,
    pub macro_args: Vec<&'s str>,
}

#[derive(Debug, PartialEq)]
pub struct Entity<'s> {
    pub source: &'s str,
    pub entity_name: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct LatexFragment<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct ExportSnippet<'s> {
    pub source: &'s str,
    pub backend: &'s str,
    pub contents: Option<&'s str>,
}

#[derive(Debug, PartialEq)]
pub struct FootnoteReference<'s> {
    pub source: &'s str,
    pub label: Option<&'s str>,
    pub definition: Vec<Object<'s>>,
}

#[derive(Debug, PartialEq)]
pub struct Citation<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct CitationReference<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct InlineBabelCall<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct InlineSourceBlock<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct LineBreak<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct Target<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct StatisticsCookie<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct Subscript<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct Superscript<'s> {
    pub source: &'s str,
}

#[derive(Debug, PartialEq)]
pub struct Timestamp<'s> {
    pub source: &'s str,
    pub timestamp_type: TimestampType,
    pub range_type: TimestampRangeType,
    pub start: Option<Date<'s>>,
    pub end: Option<Date<'s>>,
}

#[derive(Debug, PartialEq)]
pub enum TimestampType {
    Diary,
    Active,
    Inactive,
    ActiveRange,
    InactiveRange,
}

#[derive(Debug, PartialEq)]
pub enum TimestampRangeType {
    None,
    DateRange,
}

pub type YearInner = u16;
pub type MonthInner = u8;
pub type DayOfMonthInner = u8;
pub type HourInner = u8;
pub type MinuteInner = u8;

#[derive(Debug, PartialEq, Clone)]
pub struct Year(YearInner);

#[derive(Debug, PartialEq, Clone)]
pub struct Month(MonthInner);

#[derive(Debug, PartialEq, Clone)]
pub struct DayOfMonth(DayOfMonthInner);

#[derive(Debug, PartialEq, Clone)]
pub struct Hour(HourInner);

#[derive(Debug, PartialEq, Clone)]
pub struct Minute(MinuteInner);

impl Year {
    // TODO: Make a real error type instead of a boxed any error.
    pub fn new<'s>(source: &'s str) -> Result<Self, Box<dyn std::error::Error>> {
        let year = source.parse::<YearInner>()?;
        Ok(Year(year))
    }

    pub fn get_value(&self) -> YearInner {
        self.0
    }
}

impl Month {
    // TODO: Make a real error type instead of a boxed any error.
    pub fn new<'s>(source: &'s str) -> Result<Self, Box<dyn std::error::Error>> {
        let month = source.parse::<MonthInner>()?;
        if month < 1 || month > 12 {
            Err("Month exceeds possible range.")?;
        }
        Ok(Month(month))
    }

    pub fn get_value(&self) -> MonthInner {
        self.0
    }
}

impl DayOfMonth {
    // TODO: Make a real error type instead of a boxed any error.
    pub fn new<'s>(source: &'s str) -> Result<Self, Box<dyn std::error::Error>> {
        let day_of_month = source.parse::<DayOfMonthInner>()?;
        if day_of_month < 1 || day_of_month > 31 {
            Err("Day of month exceeds possible range.")?;
        }
        Ok(DayOfMonth(day_of_month))
    }

    pub fn get_value(&self) -> DayOfMonthInner {
        self.0
    }
}

impl Hour {
    // TODO: Make a real error type instead of a boxed any error.
    pub fn new<'s>(source: &'s str) -> Result<Self, Box<dyn std::error::Error>> {
        let hour = source.parse::<HourInner>()?;
        if hour < 1 || hour > 23 {
            Err("Hour exceeds possible range.")?;
        }
        Ok(Hour(hour))
    }

    pub fn get_value(&self) -> HourInner {
        self.0
    }
}

impl Minute {
    // TODO: Make a real error type instead of a boxed any error.
    pub fn new<'s>(source: &'s str) -> Result<Self, Box<dyn std::error::Error>> {
        let minute = source.parse::<MinuteInner>()?;
        if minute < 1 || minute > 59 {
            Err("Minute exceeds possible range.")?;
        }
        Ok(Minute(minute))
    }

    pub fn get_value(&self) -> MinuteInner {
        self.0
    }
}

#[derive(Debug, PartialEq, Clone)]
pub struct Date<'s> {
    year: Year,
    month: Month,
    day_of_month: DayOfMonth,
    day_name: Option<&'s str>,
}

impl<'s> Date<'s> {
    // TODO: Make a real error type instead of a boxed any error.
    pub fn new(
        year: Year,
        month: Month,
        day_of_month: DayOfMonth,
        day_name: Option<&'s str>,
    ) -> Result<Self, Box<dyn std::error::Error>> {
        // TODO: Does org-mode support non-gregorian calendars?
        // TODO: Do I want to validate leap year?
        match (month.get_value(), day_of_month.get_value()) {
            (1, 1..=31) => {}
            (2, 1..=29) => {}
            (3, 1..=31) => {}
            (4, 1..=30) => {}
            (5, 1..=31) => {}
            (6, 1..=30) => {}
            (7, 1..=31) => {}
            (8, 1..=31) => {}
            (9, 1..=30) => {}
            (10, 1..=31) => {}
            (11, 1..=30) => {}
            (12, 1..=31) => {}
            _ => Err("Invalid day of month for the month.")?,
        };
        Ok(Date {
            year,
            month,
            day_of_month,
            day_name,
        })
    }

    pub fn get_year(&self) -> &Year {
        &self.year
    }

    pub fn get_month(&self) -> &Month {
        &self.month
    }

    pub fn get_day_of_month(&self) -> &DayOfMonth {
        &self.day_of_month
    }

    pub fn get_day_name(&self) -> Option<&'s str> {
        self.day_name
    }
}

#[derive(Debug, PartialEq, Clone)]
pub struct Time<'s> {
    hour: Hour,
    minute: Minute,
    postfix: Option<&'s str>,
}

impl<'s> Time<'s> {
    // TODO: Make a real error type instead of a boxed any error.
    pub fn new(
        hour: Hour,
        minute: Minute,
        postfix: Option<&'s str>,
    ) -> Result<Self, Box<dyn std::error::Error>> {
        Ok(Time {
            hour,
            minute,
            postfix,
        })
    }

    pub fn get_hour(&self) -> &Hour {
        &self.hour
    }

    pub fn get_minute(&self) -> &Minute {
        &self.minute
    }

    pub fn get_postfix(&self) -> Option<&'s str> {
        self.postfix
    }
}

impl<'s> GetStandardProperties<'s> for Object<'s> {
    fn get_standard_properties<'b>(&'b self) -> &'b dyn StandardProperties<'s> {
        match self {
            Object::Bold(inner) => inner,
            Object::Italic(inner) => inner,
            Object::Underline(inner) => inner,
            Object::StrikeThrough(inner) => inner,
            Object::Code(inner) => inner,
            Object::Verbatim(inner) => inner,
            Object::PlainText(inner) => inner,
            Object::RegularLink(inner) => inner,
            Object::RadioLink(inner) => inner,
            Object::RadioTarget(inner) => inner,
            Object::PlainLink(inner) => inner,
            Object::AngleLink(inner) => inner,
            Object::OrgMacro(inner) => inner,
            Object::Entity(inner) => inner,
            Object::LatexFragment(inner) => inner,
            Object::ExportSnippet(inner) => inner,
            Object::FootnoteReference(inner) => inner,
            Object::Citation(inner) => inner,
            Object::CitationReference(inner) => inner,
            Object::InlineBabelCall(inner) => inner,
            Object::InlineSourceBlock(inner) => inner,
            Object::LineBreak(inner) => inner,
            Object::Target(inner) => inner,
            Object::StatisticsCookie(inner) => inner,
            Object::Subscript(inner) => inner,
            Object::Superscript(inner) => inner,
            Object::Timestamp(inner) => inner,
        }
    }
}

impl<'s> StandardProperties<'s> for Bold<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Italic<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Underline<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for StrikeThrough<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Code<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Verbatim<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for RegularLink<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for RadioLink<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for RadioTarget<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for PlainLink<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for AngleLink<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for OrgMacro<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Entity<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for LatexFragment<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for ExportSnippet<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for FootnoteReference<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Citation<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for CitationReference<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for InlineBabelCall<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for InlineSourceBlock<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for LineBreak<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Target<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for StatisticsCookie<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Subscript<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Superscript<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Timestamp<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for PlainText<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> Timestamp<'s> {
    pub fn get_raw_value(&self) -> &'s str {
        self.source.trim_end()
    }
}
